package ActiveState::Config::Simple;

use strict;
use base qw(Exporter);
our @EXPORT_OK = qw(config_parse config_parse_file);

use Carp qw(croak);
use File::Glob qw(:glob);
use ActiveState::UTF8 qw(decode_utf8);

sub config_parse_file {
    my ($file, $cnf, %opt) = @_;
    open(my $fh, "<", $file) or return undef;
    return config_parse(do {local $/; scalar <$fh>}, $file, $cnf, %opt);
}

sub config_parse {
    my($text, $file, $cnf, %opt) = @_;
    $cnf ||= {};
    $text = '' unless defined $text;

    # Decode the utf8 file text to iso-8859-1 by default.
    $text = decode_utf8($text) 
	unless $opt{text_encoding} && $opt{text_encoding} eq "utf8";
    my @stack;
    my $dir;
    my $multiline;
    foreach (split/\n/, $text) {
        if (defined $multiline) {
            if (s/\Q$multiline->{tag}\E$//) {
                # end of the multiline string
                $multiline->{val} .= $_;
                $cnf->{$multiline->{key}} = $multiline->{val};
                $multiline = undef;
            }
            else {
                $multiline->{val} .= "$_\n";
            }
            next;
        }
	next if /^\#/;
	s/^\s+//;
	next if /^$/;
	next if /^\#/;
	if (s/^<//) {
	    if (/^(\/?)([a-zA-Z_\xC0-\xD6\xD8-\xF6\xF8-\xFF][-.:\w\xC0-\xD6\xD8-\xF6\xF8-\xFF]*)\s*([-.:\w\xC0-\xD6\xD8-\xF6\xF8-\xFF]*)>\s*(?:\#.*)?$/) {
		if ($1) {
		    # end block
		    croak("Bad end marker") if length $3;
		    my $tag = $2;
		    #print "END $tag\n";
		    my $frame = pop(@stack) || croak("End marker underflow");
		    croak("Unmatched end marker") unless $frame->[1] eq $tag;
		    my $old_cnf = $cnf;
		    $cnf = $frame->[0];
		    if (my $list = $old_cnf->{'#list'}) {
			die "Both simple values an key/value pairs in $tag section"
			    if keys(%$old_cnf) > 1;
			$cnf->{$tag} = $list;
		    }
		}
		else {
		    # begin block
		    my($tag1, $tag2) = ($2, $3);
		    #print "START $tag1 $tag2\n";
		    my $new_cnf = {};
		    if (ref($cnf->{$tag1}) eq 'ARRAY') {
			croak("Not a hash") if length $tag2;
			push(@{$cnf->{$tag1}}, $new_cnf);
		    }
		    elsif (exists $cnf->{$tag1}) {
			if (length $tag2) {
			    $cnf->{$tag1}{$tag2} = $new_cnf;
			}
			else {
			    $cnf->{$tag1} = [$cnf->{$tag1}, $new_cnf];
			}
		    }
		    else {
			if (length $tag2) {
			    $cnf->{$tag1}{$tag2} = $new_cnf;
			}
			else {
			    $cnf->{$tag1} = $new_cnf;
			}
		    }
		    push(@stack, [$cnf, $tag1]);
		    $cnf = $new_cnf;
		}
	    }
	    else {
		chomp;
		croak("Unrecognized block '<$_'");
	    }
	}
	elsif (/^!\s*(\S+)\s*(.*)/) {
	    my($directive, $arg) = ($1, $2);
	    if ($directive eq "include") {
		croak("Can't include '$arg' without a root file") unless $file;
		#print "INCLUDE $arg\n";
		unless ($arg =~ m,^/,) {
		    unless ($dir) {
			require File::Basename;
			$dir = File::Basename::dirname($file);
		    }
		    $arg = join("/", $dir, $arg);
		}
		my @inc = bsd_glob($arg);
		#print "    \\___ @inc\n";
		config_parse_file($_, $cnf, %opt) for @inc;
	    }
	    else {
		croak("Unrecognized directive '$directive'");
	    }
	}
	else {
	    # should be a variable then
	    #   foo = bar
	    #   bar
	    if (s/^([-\w\xC0-\xD6\xD8-\xF6\xF8-\xFF]+)\s*=\s*//) {
		my $key = $1;
		chomp;
		my $val = _val($_);
                if (!defined $val and s/^(['"])//) {
                    my $tag = $1;
                    # add a newline if there is text on this line already
                    $_ .= "\n" if length $_;
                    $multiline = { key => $key,
                                   tag => $tag,
                                   val => $_,
                                 };
                    next;
                }

		if (ref($cnf->{$key}) eq "ARRAY") {
		    push(@{$cnf->{$key}}, $val);
		}
		elsif (exists $cnf->{$key}) {
		    $cnf->{$key} = [$cnf->{$key}, $val];
		}
		else {
		    $cnf->{$key} = $val;
		}
	    }
	    else {
		chomp;
		push(@{$cnf->{'#list'}}, _val($_));
	    }
	}
    }
    if (defined $multiline) {
        croak "Unterminated key $multiline->{key}";
    }

    if (@stack) {
	my @open_tags = map $_->[1], @stack;
	if (@open_tags == 1) {
	    croak("Unclosed group <$open_tags[0]>");
	}
	else {
	    @open_tags = map "<$_>", @open_tags;
	    my $last_tag = pop(@open_tags);
	    croak("Unclosed groups " . join(" and ", join(", ", @open_tags),
					             $last_tag));
	}
    }

    return $cnf;
}


# will return undef at the start of multi-line values
sub _val  {
    my $val = shift;
    if ($val =~ /^\"([^\"\\]*(?:\\.[^\"\\]*)*)\"\s*(?:\#.*)?$/) {
	# double quoted string
	$val = $1;
	while ($val =~ /\\(.)/g) {
	    my $c = $1;
	    croak("Bad escape \\$c in \"$val\"") if $c =~ m/[^\\\"]/;
	}
	$val =~ s/\\(.)/$1/g;
    }
    elsif ($val =~ /^\'([^\'\\]*(?:\\.[^\'\\]*)*)\'\s*(?:\#.*)?$/) {
	# single quoted string
	$val = $1;
	while ($val =~ /\\(.)/g) {
	    my $c = $1;
	    croak("Bad escape \\$c in '$val'") if $c =~ m/[^\\\']/;
	}
	$val =~ s/\\(.)/$1/g;
    }
    elsif ($val =~ /^[\'\"]/) {
	return undef;
    }
    elsif ($val =~ /^!?\s*\//) {
	# regular expression (do we really want this?)
	if ($val =~ m,[^/]*/[^/\\]*(?:\\.[^/\\]*)*/[^\#]*,) {
	    if ($+[0] != length($val)) {
		substr($val, $+[0]) = "";
	    }
	    $val =~ s/\s+$//;
	}
	elsif ($val =~ m,([^/]*/[^\#]*),) {
	    $val = $1;
	    $val =~ s/\s+$//;
	}
	else {
	    croak("Bad RE value '$val'");
	}
    }
    else {
	# bareword
	$val =~ s/\#.*//;
	$val =~ s/\s+$//;
	$val = undef if $val eq "__DEFAULT__";
    }
    return $val;
}

1;

__END__

=head1 NAME

ActiveState::Config::Simple - Simple-minded configuration file parser

=head1 SYNOPSIS

  use ActiveState::Config::Simple qw(config_parse_file);

  my $conf = config_parse_file("foo.conf");

  $conf = config_parse_file("bar.conf", $conf);
  
  my $cf_utf8 = config_parse_file("baz.conf", undef, text_encoding => 'utf8');
  
=head1 DESCRIPTION

This module exports a single function called config_parse_file() that
parse the configuration file using the same rules as
ActiveState::Config and return a hash reference similar to what
ActiveState::Config's as_hash method returns.

ActiveState::Config::Simple is not able to run validators on the
configuration values, but it runs faster.

=head2 config_parse_file(file[, hash, options])

C<file> is the config file to be parsed. C<hash> is a hash-ref where the 
parsed config data structure will be loaded. If this is undef a new hash-ref 
will be created for you. C<options> is a list of key/value option pairs. 
The only available option is C<text_encoding>. This controls how the parsed 
text is encoded in the returned data structure. C<text_encoding> can be 
C<iso-8859-1> (the default) or C<utf8>.

NOTE: The standard ActiveState::Config interface does not support the 
C<text_encoding> option. Do not take utf-8 encoded strings and use them to 
save data using the standard ActiveState::Config interface. The standard 
interface expects that input data is encoded as iso-8859-1.

=cut
