package ActiveState::Config::Tokenizer;

use strict;
use ActiveState::UTF8 qw(encode_utf8);

sub new {
    my ($cls, $type, $arg, @opts) = @_;

    # Support passing a full classname:
    return $type->new($arg, @opts) if $type =~ /::/;

    # Support passing in a subroutine reference:
    return $type->($arg, @opts) if ref($type) eq 'CODE';

    # Support passing in a shorthand for a builtin tokenizer:
    my $delegate = "ActiveState::Config::Tokenizer::\u\L$type";
    eval {
	# Check if it's already defined (avoid requiring a non-file)
	#return 1 if $delegate->can('new');

	# Try to require it and try again
	eval "require $delegate";
	die $@ if $@;
	return 1 if $delegate->can('new');

	# Obviously a bogus entry
	die "can't find $delegate->new()";
    };
    die "Can't create '$type' Tokenizer: $@" if $@;
    return $delegate->new($arg, @opts);
}

# The following methods must be provided by each Tokenizer subclass:
sub savebuf { die }	# get or save the current working buffer
sub readline { die }	# read a single line from the source, and return it
sub lineno { die }	# current line number
sub filename { die }	# the filename (could be a non-file, of course)
sub dir { die }		# the directory containing the file
sub unlock { 1 }	# unlock any locked files
sub close { 1 }		# close any opened files
sub cansave { 0 }	# can we save to this?

#=============================================================================
# Tokenizing logic:
#=============================================================================

my ($group_tag_rx, $group_name_rx, $opt_name_rx);

# Parse and return the next token from the stream:
sub next {
    my $o = shift;
    my $orig = $o->savebuf;
    my $bol = 0;
    my ($lead, $trail);
    $orig =~ s/^(\s*)//; $lead  = $1;
    $orig =~ s/(\s*)$//; $trail = $1;
    while (not length $orig) {
	$orig = $o->readline;
	$bol = 1;
	return unless defined $orig;
	chomp $orig;
	$orig =~ s/^(\s*)//; $lead  = $1;
	$orig =~ s/(\s*)$//; $trail = $1;
    }

    $group_tag_rx  ||= qr/[a-zA-Z\xC0-\xD6\xD8-\xF6\xF8-\xFF_][-\w\xC0-\xD6\xD8-\xF6\xF8-\xFF]*/;
    $group_name_rx ||= qr/[a-zA-Z\xC0-\xD6\xD8-\xF6\xF8-\xFF_][-:\w\xC0-\xD6\xD8-\xF6\xF8-\xFF]*/;
    $opt_name_rx   ||= qr/[a-zA-Z\xC0-\xD6\xD8-\xF6\xF8-\xFF_][-\w\xC0-\xD6\xD8-\xF6\xF8-\xFF]*/;

    my %token;
    if ($orig =~ s#^($opt_name_rx)\s*=\s*(\S)#$2#) {
	my $name = $1;
	my ($value, $type) = _lex_value(\$orig);
	defined $value or die "parse error.\n";
	%token = (
	    type => 'NAMEVALUE',
	    name => $name,
	    value => $value,
	    vtype => $type,
	    special => $type eq 'special',
	);
    }
    elsif ($orig =~ s/^(#.*)//) {
	my $txt = $1;
	if ($bol
		and !$lead
		and $o->lineno() == 1
		and $txt =~ s/^#\%ASCFG-(\S+)//) {
	    # We have ourselves a comment:
	    %token = (
		type => 'VERSION',
		value => $1
	    );
	}
	else {
	    my $type = $bol ? "BLOCK_COMMENT" : "LINE_COMMENT";
	    $txt .= "\n" if $type eq 'BLOCK_COMMENT';
	    %token = (
		type => $type,
		value => $txt,
	    );
	}
    }
    elsif ($orig =~ s#^<(/?)($group_tag_rx)##) {
	my ($close, $type) = ($1, $2);
	if ($close) {
	    %token = (
		type => 'CLOSE',
		group => $type,
	    );
	}
	else {
	    my $name = ($orig =~ s#^\s+($group_name_rx)\s*##) ? $1 : '';
	    %token = (
		type => 'OPEN',
		group => $type,
		$name ? (name => $name) : (),
	    );
	}
	die "illegal tag '<$close$type$orig'.\n" unless $orig =~ s#^>\s*##;
    }
    else {
	my ($directive);
	if ($orig =~ s/^!(\w+)\s+//) {
	    $directive = $1;
	}
	my ($value, $type) = _lex_value(\$orig);
	defined $value or die "parse error.\n";
	%token = (
	    $directive ? (
		type => 'DIRECTIVE',
		name => $directive,
	    ) : (
		type => 'VALUE',
		vtype => $type,
	    ),
	    value => $value,
	);
    }
    $o->savebuf($orig); # save buffer
    return \%token;
}

sub _escape {
    my ($string, $i, $raw, $q, $on_unknown_escape, $keep_esc) = @_;
    my $c = substr($raw, $$i + 1, 1);

    if ($c eq '\\') {
	$$string .= '\\' if $keep_esc;
	$$string .= '\\';
	++$$i;
    }
    elsif ($c eq $q) {
	$$string .= '\\' if $keep_esc;
	$$string .= $q;
	++$$i;
    }
    elsif ($on_unknown_escape eq 'passthru') {
	$$string .= "\\$c";
	++$$i;
    }
    elsif ($on_unknown_escape eq 'die') {
	die "illegal escape sequence \\$c";
    }
}

sub _parse_quoted {
    my $raw = shift;
    my $quote = shift;

    my $i = 1;
    my $string = '';
    my $c;
    while ($i < length($raw) and ($c = substr($raw, $i, 1)) ne $quote) {
	if ($c eq '\\') {
	    _escape(\$string, \$i, $raw, $quote, @_);
	}
	else {
	    $string .= substr($raw, $i, 1);
	}
	++$i;
    }
    ($string, $i);
}

# Removes a value from the beginning of the working buffer. Handles
# single-quoted, double-quoted, and unquoted values
sub _lex_value {
    my $work = shift;
    my ($value, $type, $copy);

  TEST:
    {
	# Quoted string:
	if ($$work =~ /^(["'])/) {
	    $copy = $$work;
	    my $q = $1;
	    my $chars;
	    ($value, $chars) = _parse_quoted($copy, $q, 'die');
	    die "invalid quoted string.\n" unless substr($copy, $chars, 1) eq $q;
	    substr($copy, 0, $chars + 1) = '';
	    if ($copy =~ /^\s*(?:#|$)/) {
		$type = $q;
		last TEST;
	    }
	}

	# A regular expression:
	if ($$work =~ m#^!?\s*/#) {
	    ($copy = $$work) =~ s#^(!?\s*)## || die;
	    my $negate = $1;
	    my $chars;
	    ($value, $chars) = _parse_quoted($copy, '/', 'passthru', 1);
	    if (substr($copy, $chars, 1) eq '/') {
		substr($copy, 0, $chars + 1) = '';
		substr($value, 0, 0, '/');
		$value .= '/';
		$value .= $1 if $copy =~ s/^([a-z]+)//;
		substr($value, 0, 0, $negate);
		if ($copy =~ /^\s*(?:#|$)/) {
		    $type = '/';
		    last TEST;
		}
	    }
	}

	# Unquoted value:
	my $tmp;
	$copy = $$work;
	($value, $tmp) = split /#/, $copy, 2;
	$value =~ s/^\s*//;
	$value =~ s/\s*$//;
	$copy = $tmp ? "#$tmp" : '';
	$type = ($value =~ /^__/ ? 'special' : '');
    }
    $$work = $copy;
    wantarray ? ($value, $type) : $value;
}

#=============================================================================
# Emitter
#=============================================================================
package ActiveState::Config::Emitter;

sub new {
    my $cls = shift;
    my ($node, $parser, $is_utf8) = @_;
    my $self = bless {
	node => $node,
	parser => $parser,
        is_utf8 => $is_utf8,
    }, ref($cls) || $cls;
}

# Decide the optimal output format for a scalar. If it contains no special
# characters, emit it verbatim. If it contains quotes, use the opposite
# quotes.
sub quote {
    my ($val, $hint) = @_;
    my $emit_type = $hint->{vtype};
    unless ($emit_type) {
	if ($val =~ /^!?\s*\//) {
	    $emit_type = '/';
	}
	elsif ($val =~ /[#"]/ and $val !~ /'/) {
	    $emit_type = "'";
	}
	elsif ($val =~ /[#']/) {
	    $emit_type = '"';
	}
	elsif (length $val == 0) {
	    $emit_type = '"';
	}
	else {
	    $emit_type = '';
	}
    }

    if ($emit_type eq "'") {
	$val =~ s/\\/\\\\/g;
	$val =~ s/'/\\'/g;
	return qq{'$val'};
    }
    elsif ($emit_type eq '"') {
	$val =~ s/\\/\\\\/g;
	$val =~ s/"/\\"/g;
	return qq{"$val"};
    }
    # do nothing: both '/' and '' emit_types need no escaping.
    $val;
}

sub CTX_OPTION        { 0 }
sub CTX_GROUP_OPEN()  { 1 }
sub CTX_GROUP_CLOSE() { 2 }

my %info;
sub walk {
    my ($o, $cb, $n) = @_;
    my $bcom = $n->{block_comment};
    my $lcom = $n->{line_comment};
    my $hint = $n->{__emit__};
    $bcom = '' unless defined $bcom;
    $lcom = '' unless defined $lcom;

    # group node
    if (defined $n->{children} or defined $n->{groupclass}) {
	my @bits;
	push @bits, $n->{groupclass} if defined $n->{groupclass};
	push @bits, $n->{name} if defined $n->{name} && length $n->{name};
	my $open = "<@bits>";
	$open .= "\t$lcom" if $lcom;
	$open .= "\n";
	%info = (
	    open	=> $open,
	    bcom	=> $bcom,
	);
	$cb->(CTX_GROUP_OPEN, \%info, $n) if $n->{path} ne '';
	for my $ch (@{$n->{children}}) {
	    $o->walk($cb, $ch);
	}
	my $close = "</$bits[0]>\n";
	%info = (close => $close);
	$cb->(CTX_GROUP_CLOSE, \%info, $n) if $n->{path} ne ''
    }
    # value node
    else {
	my ($name, $v, $d) = @$n{qw(name value default)};
	my ($value, $open, $close, @lines);
	if (defined $v) {
	    if (ref($v) eq 'CODE') {
		$value = $n->value;
	    }
	    else {
		$value = $v;
	    }
	    $value = $n->_format($value);
	}
	else {
	    # either the node didn't show up at all, or it was just assigned
	    # __DEFAULT__, and didn't have any value assigned. If the node
	    # didn't show up at all, we don't want to write it out again.
	    if ($n->{type} eq 'placeholder') {
		if ($hint->{directive}) {
		    @lines = ("\n", "$hint->{directive}\n");
		    goto EMIT_LINES;
		}
		elsif ($hint->{block_comment_only}) {
		    goto EMIT_LINES;
		}
		elsif ($hint->{infile}) {
		    goto VALUIZE_DEFAULT;
		}
		else { return }
	    }
	    else {
		VALUIZE_DEFAULT:
		# if it didn't show up from the file, don't bother outputting
		# it, since it will just be autovivified again next time.
		if (!$hint->{infile}) { return }
		elsif (ref($d) eq 'ARRAY') {
		    $value = $d;
		}
		else {
		    $value = '__DEFAULT__';
		}
	    }
	}

	if (ref($value) eq 'ARRAY') {
	    $lcom = [$lcom] if ref($lcom) ne 'ARRAY';
	    if (!$hint->{listtype} or $hint->{listtype} ne 'repeated') {
		$open = "<$n->{name}>";
		$open .= "\n";
		$close = "</$n->{name}>";
		$close .= "\n";
		for my $i (0 .. $#$value) {
		    $lines[$i] = quote($value->[$i], $hint);
		    $lines[$i] .= "\t$lcom->[$i]"
			if defined $lcom->[$i] && length $lcom->[$i];
		    $lines[$i] .= "\n";
		}
	    }
	    else {
		for my $i (0 .. $#$value) {
		    $lines[$i] = "$name = " . quote($value->[$i], $hint);
		    $lines[$i] .= "\t$lcom->[$i]"
			if defined $lcom->[$i] && length $lcom->[$i];
		    $lines[$i] .= "\n";
		}
	    }
	}
	else {
	    $lines[0] = "$name = " . quote($value, $hint);
	    $lines[0] .= "\t$lcom" if defined $lcom and length $lcom;
	    $lines[0] .= "\n";
	}

      EMIT_LINES:

	if ($open) {
	    %info = (
		bcom	=> $bcom,
		open	=> $open,
	    );
	    $cb->(CTX_GROUP_OPEN, \%info, $n);
	    %info = (bcom => '', lines => \@lines);
	    $cb->(CTX_OPTION, \%info, $n);
	    %info = (close => $close);
	    $cb->(CTX_GROUP_CLOSE, \%info, $n);
	}
	else {
	    %info = (
		bcom	=> $bcom,
		lines	=> \@lines,
	    );
	    $cb->(CTX_OPTION, \%info, $n);
	}
    }
}

# Return a string representation of the tree rooted at $o->{node}.
sub serialize {
    my $o = shift;
    my $txt = '';
    my $level = 0;
    my $step  = ' ' x 4; # four spaces
    my $sep = '';
    $o->walk(sub {
	my ($ctx, $info) = @_;
	if ($ctx eq CTX_GROUP_OPEN) {
	    my $ind = $step x $level;
	    (my $bcom = $info->{bcom}) =~ s#^\s*#$ind#mg;
	    $txt .= $sep . $bcom if $info->{bcom};
	    $txt .= $ind . $info->{open};
	    $level++;
	}
	elsif ($ctx eq CTX_GROUP_CLOSE) {
	    $level--;
	    $txt .= ($step x $level) . $info->{close};
	}
	elsif ($ctx eq CTX_OPTION) {
	    my $ind = $step x $level;
	    (my $bcom = $info->{bcom}) =~ s#^\s*#$ind#mg;
	    $txt .= $sep . $bcom if $info->{bcom};
	    for my $line (@{$info->{lines}}) {
		$txt .= $ind . $line;
	    }
	}
	$sep = "\n";
    }, $o->{node});
    $txt;
}

# local print function to encapsulate utf8_encoding.
sub _print {
    my ($o, $fh, $str) = @_;
    if ($o->{is_utf8}) {
        print $fh $str;
    } else {
        print $fh ActiveState::UTF8::encode_utf8($str);
    }
}

# Walk through the tree rooted at $o->{node}, searching for those which need
# saving. For nodes whose tokenizer doesn't exist, we look for a sibling node
# whose tokenizer *does* exist, and write to the same file.
sub save {
    my $o = shift;
    my ($toke, $file, $fh);
    my $step  = ' ' x 4; # four spaces
    my %seen;
    $o->walk(sub {
	my ($ctx, $info, $node) = @_;
	# If the node has no tokenizer defined, use the current one. This
	# should cover most cases?
	if (!$toke or $node->{tokenizer} and $toke ne $node->{tokenizer}) {
	    $toke = $node->{tokenizer};
	    $file = undef;
	    $fh = undef;
	}

	if (!$file or !$fh) {
	    return unless defined $toke
		and eval { $toke->isa("ActiveState::Config::Tokenizer") };
	    return unless $toke->{dirty} and $toke->cansave;
	    $file = $toke->filename;
	    $fh = $toke->prepare_save;
	    unless ($seen{$file}{hits}++) {
		$seen{$file}{fh} = $fh;
		$seen{$file}{level} = 0;
		$seen{$file}{sep} = ''; # no leading blank
		$seen{$file}{toke} = $toke;
	    }
	}
	return unless $toke->{dirty};

	# Now do the regular song and dance
	if ($ctx eq CTX_GROUP_OPEN) {
	    my $ind = $step x $seen{$file}{level};
	    (my $bcom = $info->{bcom}) =~ s#^\s*#$ind#mg;
	    if ($info->{bcom}) {
		$o->_print($fh, $seen{$file}{sep} . $bcom)
                    or die "can't save $file: $!";
	    }
            $o->_print($fh, $ind . $info->{open}) or die "can't save $file: $!";
	    $seen{$file}{level}++;
	}
	elsif ($ctx eq CTX_GROUP_CLOSE) {
	    $seen{$file}{level}--;
            $o->_print($fh, ($step x $seen{$file}{level}) . $info->{close})
                or die "can't save $file: $!";
	}
	elsif ($ctx eq CTX_OPTION) {
	    my $ind = $step x $seen{$file}{level};
	    (my $bcom = $info->{bcom}) =~ s#^\s*#$ind#mg;
	    if ($info->{bcom}) {
                $o->_print($fh, $seen{$file}{sep} . $bcom)
                    or die "can't save $file: $!";
	    }
	    for my $line (@{$info->{lines}}) {
                $o->_print($fh, $ind . $line) or die "can't save $file: $!";
	    }
	}
	$seen{$file}{sep} = "\n";
    }, $o->{node});
    for my $file (keys %seen) {
	$seen{$file}{toke}->commit_save;
	$seen{$file}{toke}{dirty} = 0;
    }
    # If you delete all nodes from a file, it won't be %seen. So walk through
    # the list of tokenizers, and make sure to truncate all files which
    # haven't been touched.
    for my $toke (@{$o->{parser}{tokenizers}}) {
	next unless $toke->{dirty};
	my $file = $toke->filename;
	next if $seen{$file};
	$toke->prepare_save;
	$toke->commit_save;
    }
}

1;
