package ActiveState::Config::Tokenizer::File;

use strict;
use ActiveState::File::Atomic;
use base qw(ActiveState::Config::Tokenizer);
use ActiveState::UTF8 qw(decode_utf8);

sub new {
    my ($cls, $file, %opts) = @_;
    my $o = bless {
	file => $file,
	line => 0,
	opts => \%opts,
    }, ref($cls) || $cls;
    $o->open;
    $o;
}

sub open {
    my $o = shift;
    my $f = $o->{file};
    my $rw = $o->{opts}{writable} ? 'r/w' : 'r';
    my %opts = (
	writable	=> 0,
	timeout		=> 0,		# wait forever
	backup_ext	=> undef,	# no backups,
	rotate		=> 0,
	%{$o->{opts}},
    );

    # AS::File::Atomic croaks if you specify an undefined value for
    # backup_ext.
    delete $opts{backup_ext} unless defined $opts{backup_ext};
    # the same applies for is_utf8.
    delete $opts{is_utf8};
    
    $o->{__file__} = ActiveState::File::Atomic->new($f, %opts);
}

sub close {
    my $o = shift;
    delete @$o{qw(wfd __file__)}; # reverts, closes, and unlocks
    1;
}

sub DESTROY {
    my $o = shift;
    $o->close;
}

sub savebuf {
    my $o = shift;
    if (@_) {
	$o->{savebuf} = shift;
    }
    defined $o->{savebuf} ? $o->{savebuf} : '';
}

sub readline {
    my $o = shift;
    ++$o->{line};
    my $l = $o->{__file__}->readline;
    if (!defined $l and !$o->{opts}{writable}) {
	$o->close;
    }

    if (defined $l) {
	if ($o->{opts}{is_utf8}) {
	    return $l;
	} else {
	    return decode_utf8($l);
	}
    } else {
	return $l;
    }
}

sub lineno {
    my $o = shift;
    $o->{line};
}

sub filename {
    my $o = shift;
    $o->{file};
}

sub dir {
    my $o = shift;
    use Cwd qw(cwd);
    use File::Basename qw(dirname);
    my $d = dirname($o->{file});
    $d = cwd . "/$d" unless $d =~ m#^/#;
    $d;
}

#=============================================================================
# Support for saving. Note: after calling save(), it is an error to change
# anything in the parser, as all files will be closed and unlocked.
#=============================================================================
sub cansave { 1 }

sub prepare_save {
    my $o = shift;
    die "can't save to file $o->{file}: was not opened writable"
	unless $o->{opts}{writable};
    return $o->{wfd} ||= $o->{__file__}->tempfile;
}

sub commit_save {
    my $o = shift;
    delete $o->{wfd};  # commit_tempfile owns the file and will close it
    $o->{__file__}->commit_tempfile;
    $o->close;
}

1;
